import logging
from collections.abc import Coroutine, Iterable, Mapping
from concurrent.futures import Future
from typing import Any, cast, overload

import anyio
import anyio.to_thread
import confluent_kafka
import confluent_kafka.admin

from ._utils import FuturesDict, make_kwargs, make_sync_config, to_dict, to_list, to_set, wrap_concurrent_future


class AdminClient:
    def __init__(self, config: Mapping[str, Any], *, logger: logging.Logger | None = None) -> None:
        self.sync: confluent_kafka.admin.AdminClient = confluent_kafka.admin.AdminClient(
            make_sync_config(config),  # pyright: ignore[reportArgumentType]
            logger=logger,
        )

    def create_topics(
        self,
        new_topics: Iterable[confluent_kafka.admin.NewTopic],
        *,
        operation_timeout: float | None = None,
        request_timeout: float | None = None,
        validate_only: bool = False,
    ) -> FuturesDict[str, None]:
        return FuturesDict[str, None].from_concurrent_futures(
            self.sync.create_topics(
                to_list(new_topics),
                validate_only=validate_only,
                **make_kwargs(operation_timeout=operation_timeout, request_timeout=request_timeout),
            )
        )

    def delete_topics(
        self,
        topics: Iterable[str],
        *,
        operation_timeout: float | None = None,
        request_timeout: float | None = None,
    ) -> FuturesDict[str, None]:
        return FuturesDict[str, None].from_concurrent_futures(
            self.sync.delete_topics(
                to_list(topics), **make_kwargs(operation_timeout=operation_timeout, request_timeout=request_timeout)
            )
        )

    async def list_topics(
        self, topic: str | None = None, timeout: float | None = None
    ) -> confluent_kafka.admin.ClusterMetadata:
        return await anyio.to_thread.run_sync(lambda: self.sync.list_topics(topic, **make_kwargs(timeout=timeout)))

    def create_partitions(
        self,
        new_partitions: Iterable[confluent_kafka.admin.NewPartitions],
        *,
        operation_timeout: float | None = None,
        request_timeout: float | None = None,
        validate_only: bool = False,
    ) -> FuturesDict[str, None]:
        return FuturesDict[str, None].from_concurrent_futures(
            self.sync.create_partitions(
                to_list(new_partitions),
                validate_only=validate_only,
                **make_kwargs(operation_timeout=operation_timeout, request_timeout=request_timeout),
            )
        )

    def describe_configs(
        self,
        resources: Iterable[confluent_kafka.admin.ConfigResource],
        *,
        request_timeout: float | None = None,
    ) -> FuturesDict[confluent_kafka.admin.ConfigResource, dict[str, confluent_kafka.admin.ConfigEntry]]:
        return FuturesDict[
            confluent_kafka.admin.ConfigResource, dict[str, confluent_kafka.admin.ConfigEntry]
        ].from_concurrent_futures(
            self.sync.describe_configs(to_list(resources), **make_kwargs(request_timeout=request_timeout))
        )

    def incremental_alter_configs(
        self,
        resources: Iterable[confluent_kafka.admin.ConfigResource],
        *,
        request_timeout: float | None = None,
        validate_only: bool = False,
        broker: int | None = None,
    ) -> FuturesDict[confluent_kafka.admin.ConfigResource, None]:
        return FuturesDict[confluent_kafka.admin.ConfigResource, None].from_concurrent_futures(
            self.sync.incremental_alter_configs(
                to_list(resources),
                validate_only=validate_only,
                **make_kwargs(request_timeout=request_timeout, broker=broker),
            )
        )

    def create_acls(
        self,
        acls: Iterable[confluent_kafka.admin.AclBinding],
        *,
        request_timeout: float | None = None,
    ) -> FuturesDict[confluent_kafka.admin.AclBinding, None]:
        return FuturesDict[confluent_kafka.admin.AclBinding, None].from_concurrent_futures(
            self.sync.create_acls(to_list(acls), **make_kwargs(request_timeout=request_timeout))
        )

    async def describe_acls(
        self,
        acl_binding_filter: confluent_kafka.admin.AclBindingFilter,
        *,
        request_timeout: float | None = None,
    ) -> list[confluent_kafka.admin.AclBinding]:
        return await wrap_concurrent_future(
            self.sync.describe_acls(acl_binding_filter, **make_kwargs(request_timeout=request_timeout))
        )

    def delete_acls(
        self,
        acl_binding_filters: Iterable[confluent_kafka.admin.AclBindingFilter],
        *,
        request_timeout: float | None = None,
    ) -> FuturesDict[confluent_kafka.admin.AclBindingFilter, list[confluent_kafka.admin.AclBinding]]:
        return FuturesDict[
            confluent_kafka.admin.AclBindingFilter, list[confluent_kafka.admin.AclBinding]
        ].from_concurrent_futures(
            self.sync.delete_acls(to_list(acl_binding_filters), **make_kwargs(request_timeout=request_timeout))
        )

    async def list_consumer_groups(
        self,
        *,
        request_timeout: float | None = None,
        states: Iterable[confluent_kafka.ConsumerGroupState] | None = None,
        types: Iterable[confluent_kafka.ConsumerGroupType] | None = None,
    ) -> confluent_kafka.admin.ListConsumerGroupsResult:
        return await wrap_concurrent_future(
            self.sync.list_consumer_groups(
                **make_kwargs(
                    request_timeout=request_timeout,
                    states=to_set(states) if states is not None else None,
                    types=to_set(types) if types is not None else None,
                )
            )
        )

    def describe_consumer_groups(
        self,
        group_ids: Iterable[str],
        *,
        include_authorized_operations: bool = False,
        request_timeout: float | None = None,
    ) -> FuturesDict[str, confluent_kafka.admin.ConsumerGroupDescription]:
        return FuturesDict[str, confluent_kafka.admin.ConsumerGroupDescription].from_concurrent_futures(
            self.sync.describe_consumer_groups(
                to_list(group_ids),
                include_authorized_operations=include_authorized_operations,
                **make_kwargs(request_timeout=request_timeout),
            )
        )

    def describe_topics(
        self,
        topics: confluent_kafka.TopicCollection | Iterable[str],
        *,
        include_authorized_operations: bool = False,
        request_timeout: float | None = None,
    ) -> FuturesDict[str, confluent_kafka.admin.TopicDescription]:
        return FuturesDict[str, confluent_kafka.admin.TopicDescription].from_concurrent_futures(
            self.sync.describe_topics(
                confluent_kafka.TopicCollection(to_list(topics)) if isinstance(topics, Iterable) else topics,
                include_authorized_operations=include_authorized_operations,
                **make_kwargs(request_timeout=request_timeout),
            )
        )

    async def describe_cluster(
        self,
        *,
        include_authorized_operations: bool = False,
        request_timeout: float | None = None,
    ) -> confluent_kafka.admin.DescribeClusterResult:
        return await wrap_concurrent_future(
            self.sync.describe_cluster(
                include_authorized_operations=include_authorized_operations,
                **make_kwargs(request_timeout=request_timeout),
            )
        )

    def delete_consumer_groups(
        self, group_ids: Iterable[str], *, request_timeout: float | None = None
    ) -> FuturesDict[str, None]:
        return FuturesDict[str, None].from_concurrent_futures(
            self.sync.delete_consumer_groups(to_list(group_ids), **make_kwargs(request_timeout=request_timeout))
        )

    def list_consumer_group_offsets(
        self,
        list_consumer_group_offsets_request: Iterable[confluent_kafka.ConsumerGroupTopicPartitions],
        *,
        require_stable: bool = False,
        request_timeout: float | None = None,
    ) -> FuturesDict[str, confluent_kafka.ConsumerGroupTopicPartitions]:
        return FuturesDict[str, confluent_kafka.ConsumerGroupTopicPartitions].from_concurrent_futures(
            self.sync.list_consumer_group_offsets(
                to_list(list_consumer_group_offsets_request),
                require_stable=require_stable,
                **make_kwargs(request_timeout=request_timeout),
            )
        )

    def alter_consumer_group_offsets(
        self,
        alter_consumer_group_offsets_request: Iterable[confluent_kafka.ConsumerGroupTopicPartitions],
        *,
        request_timeout: float | None = None,
    ) -> FuturesDict[confluent_kafka.ConsumerGroupTopicPartitions, confluent_kafka.ConsumerGroupTopicPartitions]:
        return FuturesDict[
            confluent_kafka.ConsumerGroupTopicPartitions, confluent_kafka.ConsumerGroupTopicPartitions
        ].from_concurrent_futures(
            self.sync.alter_consumer_group_offsets(
                to_list(alter_consumer_group_offsets_request), **make_kwargs(request_timeout=request_timeout)
            )
        )

    def set_sasl_credentials(self, username: str, password: str) -> None:
        self.sync.set_sasl_credentials(username, password)

    @overload
    async def describe_user_scram_credentials(
        self, users: None = None, *, request_timeout: float | None = None
    ) -> dict[str, confluent_kafka.admin.UserScramCredentialsDescription]: ...

    @overload
    def describe_user_scram_credentials(
        self, users: Iterable[str], *, request_timeout: float | None = None
    ) -> FuturesDict[str, confluent_kafka.admin.UserScramCredentialsDescription]: ...

    @overload
    def describe_user_scram_credentials(
        self, users: None | Iterable[str], *, request_timeout: float | None = None
    ) -> (
        Coroutine[Any, Any, dict[str, confluent_kafka.admin.UserScramCredentialsDescription]]
        | FuturesDict[str, confluent_kafka.admin.UserScramCredentialsDescription]
    ): ...

    def describe_user_scram_credentials(
        self,
        users: Iterable[str] | None = None,
        *,
        request_timeout: float | None = None,
    ) -> (
        Coroutine[Any, Any, dict[str, confluent_kafka.admin.UserScramCredentialsDescription]]
        | FuturesDict[str, confluent_kafka.admin.UserScramCredentialsDescription]
    ):
        kwargs = make_kwargs(request_timeout=request_timeout)
        if users is None:
            future = cast(
                Future[dict[str, confluent_kafka.admin.UserScramCredentialsDescription]],
                self.sync.describe_user_scram_credentials(**kwargs),
            )
            return wrap_concurrent_future(future)
        futures_dict = self.sync.describe_user_scram_credentials(to_list(users), **kwargs)
        assert isinstance(futures_dict, dict)
        return FuturesDict[str, confluent_kafka.admin.UserScramCredentialsDescription].from_concurrent_futures(
            futures_dict
        )

    def alter_user_scram_credentials(
        self,
        alterations: Iterable[confluent_kafka.admin.UserScramCredentialAlteration],
        *,
        request_timeout: float | None = None,
    ) -> FuturesDict[str, None]:
        return FuturesDict[str, None].from_concurrent_futures(
            self.sync.alter_user_scram_credentials(to_list(alterations), **make_kwargs(request_timeout=request_timeout))
        )

    def list_offsets(
        self,
        topic_partition_offsets: Mapping[confluent_kafka.TopicPartition, confluent_kafka.admin.OffsetSpec],
        *,
        isolation_level: confluent_kafka.IsolationLevel | None = None,
        request_timeout: float | None = None,
    ) -> FuturesDict[confluent_kafka.TopicPartition, confluent_kafka.admin.ListOffsetsResultInfo]:
        return FuturesDict[
            confluent_kafka.TopicPartition, confluent_kafka.admin.ListOffsetsResultInfo
        ].from_concurrent_futures(
            self.sync.list_offsets(
                to_dict(topic_partition_offsets),
                **make_kwargs(isolation_level=isolation_level, request_timeout=request_timeout),
            )
        )

    def delete_records(
        self,
        topic_partition_offsets: Iterable[confluent_kafka.TopicPartition],
        *,
        request_timeout: float | None = None,
        operation_timeout: float | None = None,
    ) -> FuturesDict[confluent_kafka.TopicPartition, confluent_kafka.admin.DeletedRecords]:
        return FuturesDict[
            confluent_kafka.TopicPartition, confluent_kafka.admin.DeletedRecords
        ].from_concurrent_futures(
            self.sync.delete_records(
                to_list(topic_partition_offsets),
                **make_kwargs(request_timeout=request_timeout, operation_timeout=operation_timeout),
            )
        )

    async def elect_leaders(
        self,
        election_type: confluent_kafka.ElectionType,
        partitions: Iterable[confluent_kafka.TopicPartition] | None = None,
        *,
        request_timeout: float | None = None,
        operation_timeout: float | None = None,
    ) -> dict[confluent_kafka.TopicPartition, confluent_kafka.KafkaException | None]:
        return await wrap_concurrent_future(
            self.sync.elect_leaders(
                election_type,
                to_list(partitions) if partitions is not None else None,
                **make_kwargs(request_timeout=request_timeout, operation_timeout=operation_timeout),
            )
        )



import time
from collections.abc import AsyncIterator
from typing import Any
from uuid import uuid1

import confluent_kafka.admin
import pytest
from lovely.pytest.docker.compose import Services

from kafka_async import AdminClient, Consumer, Producer


@pytest.fixture(scope='session', autouse=True)
def anyio_backend() -> str:
    return 'asyncio'


@pytest.fixture(scope='session')
def kafka_addr() -> str:
    return '127.0.0.1:9092'


@pytest.fixture(scope='session')
def default_config(kafka_addr: str) -> dict[str, Any]:
    return {'bootstrap.servers': kafka_addr, 'topic.metadata.refresh.interval.ms': 1000}


@pytest.fixture(scope='session', autouse=True)
def start_kafka(default_config: dict[str, Any], docker_services: Services) -> None:
    docker_services.start('kafka')
    exception = None
    for _ in range(50):
        try:
            confluent_kafka.admin.AdminClient(default_config).list_topics(timeout=10)  # pyright: ignore[reportArgumentType]
        except confluent_kafka.KafkaException as exc:
            time.sleep(0.3)
            exception = exc
        else:
            return
    raise TimeoutError from exception


@pytest.fixture
async def producer(default_config: dict[str, Any]) -> AsyncIterator[Producer]:
    async with Producer(default_config) as producer:
        yield producer


@pytest.fixture
async def admin_client(default_config: dict[str, Any]) -> AdminClient:
    return AdminClient(default_config)


@pytest.fixture
async def consumer(default_config: dict[str, Any]) -> AsyncIterator[Consumer]:
    async with Consumer({**default_config, 'group.id': str(uuid1()), 'auto.offset.reset': 'earliest'}) as consumer:
        yield consumer


@pytest.fixture
def topic() -> str:
    return str(uuid1())



import sys
from collections.abc import Awaitable, Callable, Coroutine, Iterable, Mapping
from types import TracebackType
from typing import Any, Literal, overload

import anyio
import anyio.to_thread
import confluent_kafka
import confluent_kafka.admin

from ._utils import async_to_sync, make_kwargs, make_sync_config, to_list

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Consumer:
    def __init__(self, config: Mapping[str, Any]) -> None:
        self.sync: confluent_kafka.Consumer = confluent_kafka.Consumer(
            make_sync_config(config),  # pyright: ignore[reportArgumentType]
        )

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        with anyio.CancelScope(shield=True):
            await self.close()

    def assign(self, partitions: Iterable[confluent_kafka.TopicPartition]) -> None:
        self.sync.assign(to_list(partitions))

    def assignment(self) -> list[confluent_kafka.TopicPartition]:
        return self.sync.assignment()

    async def close(self) -> None:
        await anyio.to_thread.run_sync(self.sync.close)

    @overload
    def commit(
        self,
        message: confluent_kafka.Message | None = None,
        offsets: Iterable[confluent_kafka.TopicPartition] | None = None,
        *,
        asynchronous: Literal[True] = True,
    ) -> None: ...

    @overload
    async def commit(
        self,
        message: confluent_kafka.Message | None = None,
        offsets: Iterable[confluent_kafka.TopicPartition] | None = None,
        *,
        asynchronous: Literal[False],
    ) -> list[confluent_kafka.TopicPartition]: ...

    def commit(
        self,
        message: confluent_kafka.Message | None = None,
        offsets: Iterable[confluent_kafka.TopicPartition] | None = None,
        *,
        asynchronous: bool = True,
    ) -> None | Coroutine[Any, Any, list[confluent_kafka.TopicPartition]]:
        offsets_list = to_list(offsets) if offsets is not None else None
        kwargs = make_kwargs(message=message, offsets=offsets_list)
        if asynchronous:
            return self.sync.commit(**kwargs)  # pyright: ignore[reportUnknownVariableType]

        return anyio.to_thread.run_sync(lambda: self.sync.commit(asynchronous=False, **kwargs))  # pyright: ignore[reportUnknownVariableType,reportUnknownLambdaType]

    async def committed(
        self, partitions: Iterable[confluent_kafka.TopicPartition], timeout: float | None = None
    ) -> list[confluent_kafka.TopicPartition]:
        return await anyio.to_thread.run_sync(
            lambda: self.sync.committed(to_list(partitions), **make_kwargs(timeout=timeout))
        )

    async def consume(self, num_messages: int = 1, timeout: float | None = None) -> list[confluent_kafka.Message]:
        return await anyio.to_thread.run_sync(lambda: self.sync.consume(num_messages, **make_kwargs(timeout=timeout)))

    def consumer_group_metadata(self) -> object:
        return self.sync.consumer_group_metadata()

    @overload
    def get_watermark_offsets(
        self, partition: confluent_kafka.TopicPartition, *, cached: Literal[True]
    ) -> tuple[int, int]: ...

    @overload
    async def get_watermark_offsets(
        self, partition: confluent_kafka.TopicPartition, timeout: float | None = None, *, cached: Literal[False] = False
    ) -> tuple[int, int]: ...

    def get_watermark_offsets(
        self,
        partition: confluent_kafka.TopicPartition,
        timeout: float | None = None,
        cached: bool = False,
    ) -> tuple[int, int] | Coroutine[Any, Any, tuple[int, int]]:
        kwargs = make_kwargs(timeout=timeout)
        if cached:
            return self.sync.get_watermark_offsets(partition, cached=True, **kwargs)
        return anyio.to_thread.run_sync(lambda: self.sync.get_watermark_offsets(partition, cached=False, **kwargs))

    def incremental_assign(self, partitions: Iterable[confluent_kafka.TopicPartition]) -> None:
        self.sync.incremental_assign(to_list(partitions))

    def incremental_unassign(self, partitions: Iterable[confluent_kafka.TopicPartition]) -> None:
        self.sync.incremental_unassign(to_list(partitions))

    async def list_topics(
        self, topic: str | None = None, timeout: float | None = None
    ) -> confluent_kafka.admin.ClusterMetadata:
        return await anyio.to_thread.run_sync(lambda: self.sync.list_topics(topic, **make_kwargs(timeout=timeout)))

    def memberid(self) -> str | None:
        return self.sync.memberid()

    async def offsets_for_times(
        self, partitions: Iterable[confluent_kafka.TopicPartition], timeout: float | None = None
    ) -> list[confluent_kafka.TopicPartition]:
        return await anyio.to_thread.run_sync(
            lambda: self.sync.offsets_for_times(to_list(partitions), **make_kwargs(timeout=timeout))
        )

    def pause(self, partitions: Iterable[confluent_kafka.TopicPartition]) -> None:
        self.sync.pause(to_list(partitions))

    async def poll(self, timeout: float | None = None) -> confluent_kafka.Message | None:
        return await anyio.to_thread.run_sync(lambda: self.sync.poll(**make_kwargs(timeout=timeout)))

    def position(self, partitions: Iterable[confluent_kafka.TopicPartition]) -> list[confluent_kafka.TopicPartition]:
        return self.sync.position(to_list(partitions))

    def resume(self, partitions: Iterable[confluent_kafka.TopicPartition]) -> None:
        self.sync.resume(to_list(partitions))

    def seek(self, partition: confluent_kafka.TopicPartition) -> None:
        self.sync.seek(partition)

    def set_sasl_credentials(self, username: str, password: str) -> None:
        self.sync.set_sasl_credentials(username, password)

    def store_offsets(
        self,
        message: confluent_kafka.Message | None = None,
        offsets: Iterable[confluent_kafka.TopicPartition] | None = None,
    ) -> None:
        kwargs = make_kwargs(
            message=message,
            offsets=to_list(offsets) if offsets is not None else None,
        )
        self.sync.store_offsets(**kwargs)

    def subscribe(
        self,
        topics: Iterable[str],
        *,
        on_assign: (
            Callable[[confluent_kafka.Consumer, list[confluent_kafka.TopicPartition]], Awaitable[None] | None] | None
        ) = None,
        on_revoke: (
            Callable[[confluent_kafka.Consumer, list[confluent_kafka.TopicPartition]], Awaitable[None] | None] | None
        ) = None,
        on_lost: (
            Callable[[confluent_kafka.Consumer, list[confluent_kafka.TopicPartition]], Awaitable[None] | None] | None
        ) = None,
    ) -> None:
        self.sync.subscribe(
            to_list(topics),
            **make_kwargs(
                on_assign=async_to_sync(on_assign) if on_assign else None,
                on_revoke=async_to_sync(on_revoke) if on_revoke else None,
                on_lost=async_to_sync(on_lost) if on_lost else None,
            ),
        )

    def unassign(self) -> None:
        self.sync.unassign()

    def unsubscribe(self) -> None:
        self.sync.unsubscribe()



from ._consumer import Consumer
from ._producer import Producer
from ._utils import FuturesDict
from ._version import __version__, __version_tuple__
from .admin import AdminClient

__all__ = [
    'AdminClient',
    'Consumer',
    'FuturesDict',
    'Producer',
    '__version__',
    '__version_tuple__',
]



import sys
from collections.abc import Awaitable, Callable, Iterable, Mapping, Sequence
from datetime import datetime
from types import TracebackType
from typing import Any

import anyio
import anyio.to_thread
import confluent_kafka
import confluent_kafka.admin

from kafka_async._utils import async_to_sync, make_kwargs, make_sync_config, to_list

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Producer:
    def __init__(self, config: Mapping[str, Any]) -> None:
        self.sync: confluent_kafka.Producer = confluent_kafka.Producer(
            make_sync_config(config),  # pyright: ignore[reportArgumentType]
        )

    def __len__(self) -> int:
        return len(self.sync)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        with anyio.CancelScope(shield=True):
            await self.flush()

    async def abort_transaction(self, timeout: float | None = None) -> None:
        await anyio.to_thread.run_sync(lambda: self.sync.abort_transaction(**make_kwargs(timeout=timeout)))

    def begin_transaction(self) -> None:
        self.sync.begin_transaction()

    async def commit_transaction(self, timeout: float | None = None) -> None:
        await anyio.to_thread.run_sync(lambda: self.sync.commit_transaction(**make_kwargs(timeout=timeout)))

    async def flush(self, timeout: float | None = None) -> None:
        await anyio.to_thread.run_sync(lambda: self.sync.flush(**make_kwargs(timeout=timeout)))

    async def init_transactions(self, timeout: float | None = None) -> None:
        await anyio.to_thread.run_sync(lambda: self.sync.init_transactions(**make_kwargs(timeout=timeout)))

    async def list_topics(
        self, topic: str | None = None, timeout: float | None = None
    ) -> confluent_kafka.admin.ClusterMetadata:
        return await anyio.to_thread.run_sync(lambda: self.sync.list_topics(topic, **make_kwargs(timeout=timeout)))

    async def poll(self, timeout: float | None = None) -> int:
        return await anyio.to_thread.run_sync(lambda: self.sync.poll(**make_kwargs(timeout=timeout)))

    def produce(
        self,
        topic: str,
        value: str | bytes | None = None,
        *,
        key: str | bytes | None = None,
        partition: int | None = None,
        on_delivery: (
            Callable[[confluent_kafka.KafkaError | None, confluent_kafka.Message], Awaitable[None] | None] | None
        ) = None,
        timestamp: int | datetime | None = None,
        headers: Mapping[str, str | bytes | None] | Sequence[tuple[str, str | bytes | None]] | None = None,
    ) -> None:
        self.sync.produce(
            topic=topic,
            key=key,
            value=value,
            on_delivery=async_to_sync(on_delivery) if on_delivery else None,
            headers=(
                [(k, v) for k, v in headers.items()]
                if isinstance(headers, Mapping)
                else to_list(headers)
                if isinstance(headers, Sequence)
                else None
            ),
            **make_kwargs(
                timestamp=int(timestamp.timestamp() * 1000) if isinstance(timestamp, datetime) else timestamp,
                partition=partition,
            ),
        )

    async def purge(self, in_queue: bool = True, in_flight: bool = True, blocking: bool = True) -> None:
        await anyio.to_thread.run_sync(self.sync.purge, in_queue, in_flight, blocking)

    async def send_offsets_to_transaction(
        self, positions: Iterable[confluent_kafka.TopicPartition], group_metadata: object, timeout: float | None = None
    ) -> None:
        await anyio.to_thread.run_sync(
            lambda: self.sync.send_offsets_to_transaction(
                to_list(positions), group_metadata, **make_kwargs(timeout=timeout)
            ),
        )

    def set_sasl_credentials(self, username: str, password: str) -> None:
        self.sync.set_sasl_credentials(username, password)



import sys
from collections.abc import Awaitable
from uuid import uuid1

import confluent_kafka.admin
import pytest

from kafka_async import AdminClient

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup


async def test_create_topics(admin_client: AdminClient) -> None:
    name = str(uuid1())
    result = await admin_client.create_topics([confluent_kafka.admin.NewTopic(name)])
    assert result == {name: None}


async def test_list_topics(admin_client: AdminClient) -> None:
    metadata = await admin_client.list_topics()
    assert isinstance(metadata, confluent_kafka.admin.ClusterMetadata)


async def test_describe_user_scram_credentials__no_users(admin_client: AdminClient) -> None:
    assert await admin_client.describe_user_scram_credentials() == {}


async def test_describe_user_scram_credentials__with_users_and_await_one(admin_client: AdminClient) -> None:
    result = admin_client.describe_user_scram_credentials(users=['user1'])
    assert list(result.keys()) == ['user1']
    assert isinstance(result['user1'], Awaitable)
    with pytest.raises(confluent_kafka.KafkaException):
        await result['user1']


async def test_describe_user_scram_credentials__with_users_and_await_all(admin_client: AdminClient) -> None:
    with pytest.raises(ExceptionGroup) as excinfo:
        await admin_client.describe_user_scram_credentials(users=['user1'])

    assert len(excinfo.value.exceptions) == 1
    assert isinstance(excinfo.value.exceptions[0], confluent_kafka.KafkaException)


def test_set_sasl_credentials(admin_client: AdminClient) -> None:
    admin_client.set_sasl_credentials('user', 'password')



from datetime import datetime, timezone
from uuid import uuid1

import confluent_kafka
from confluent_kafka import KafkaError, Message

from kafka_async import Consumer, Producer


async def test_produce(producer: Producer, consumer: Consumer):
    topic = str(uuid1())
    value = str(uuid1())
    date = datetime(2023, 1, 1, tzinfo=timezone.utc)

    on_delivery_err: confluent_kafka.KafkaError | None = None
    on_delivery_msg: confluent_kafka.Message | None = None

    async def on_delivery(err: KafkaError | None, msg: Message | None) -> None:
        nonlocal on_delivery_err, on_delivery_msg
        on_delivery_err = err
        on_delivery_msg = msg

    producer.produce(
        topic=topic,
        value=value,
        on_delivery=on_delivery,
        headers={'a': 'string', 'b': b'bytes', 'c': None},
        timestamp=date,
    )
    await producer.flush(timeout=10)

    assert on_delivery_err is None
    assert on_delivery_msg
    assert on_delivery_msg.topic() == topic
    assert on_delivery_msg.value() == value.encode('ascii')

    consumer.subscribe([topic])
    (msg,) = await consumer.consume(timeout=5)
    assert msg.headers() == [('a', b'string'), ('b', b'bytes'), ('c', None)]
    assert msg.timestamp() == (confluent_kafka.TIMESTAMP_CREATE_TIME, int(date.timestamp() * 1000))


async def test_purge(producer: Producer) -> None:
    topic = str(uuid1())

    async def on_delivery(err: KafkaError | None, msg: Message | None) -> None:
        assert err
        assert err.code() == confluent_kafka.KafkaError._PURGE_QUEUE

    producer.produce(topic=topic, on_delivery=on_delivery)
    await producer.purge()
    await producer.flush(timeout=10)


async def test_poll(producer: Producer) -> None:
    topic = str(uuid1())
    called = False

    async def on_delivery(err: KafkaError | None, msg: Message | None) -> None:
        nonlocal called
        called = True

    producer.produce(topic=topic, on_delivery=on_delivery)
    polled = await producer.poll(timeout=5)

    assert called
    assert polled == 1

    polled = await producer.poll(timeout=0.1)
    assert polled == 0


async def test__len__(producer: Producer) -> None:
    topic = str(uuid1())
    assert len(producer) == 0
    producer.produce(topic=topic)
    assert len(producer) == 1


async def test_list_topics__all(producer: Producer) -> None:
    topic = str(uuid1())
    producer.produce(topic=topic)
    await producer.flush(timeout=10)
    topics = await producer.list_topics()
    assert topic in topics.topics


async def test_list_topics__one(producer: Producer) -> None:
    topic = str(uuid1())
    producer.produce(topic=topic)
    await producer.flush(timeout=10)
    topics = await producer.list_topics(topic)
    assert topic in topics.topics


def test_set_sasl_credentials(producer: Producer) -> None:
    producer.set_sasl_credentials('user', 'password')



import sys
from collections.abc import Awaitable, Callable, Coroutine, Generator, Iterable, Mapping
from concurrent.futures import Future
from functools import wraps
from typing import Any, ParamSpec, TypeVar

import anyio.from_thread
from anyio.lowlevel import current_token

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from exceptiongroup import ExceptionGroup
    from typing_extensions import Self

_P = ParamSpec('_P')
_T = TypeVar('_T')
_K = TypeVar('_K')
_V = TypeVar('_V')


def async_to_sync(
    func: Callable[_P, Awaitable[_T] | _T],
) -> Callable[_P, _T]:
    @wraps(func)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        result = func(*args, **kwargs)
        if isinstance(result, Awaitable):
            return anyio.from_thread.run(lambda: result)  # pyright: ignore[reportUnknownLambdaType]
        return result

    return wrapper


async def wrap_concurrent_future(future: Future[_T]) -> _T:
    if future.done():
        return future.result()
    event = anyio.Event()
    event_loop_token = current_token()
    future.add_done_callback(lambda _: anyio.from_thread.run_sync(event.set, token=event_loop_token))
    await event.wait()
    return future.result()


class FuturesDict(dict[_K, Coroutine[Any, Any, _V]]):
    """
    A dictionary which values are awaitable objects.
    It can be awaited to get a dictionary of results.

    .. code-block:: python

       async def foo(x: int) -> int:
           return x ** 2

       data = FuturesDict({1: foo(1), 2: foo(2)})
       assert await data == {1: 1, 2: 4}
    """

    def __await__(self) -> Generator[Any, Any, dict[_K, _V]]:
        return self.__await().__await__()

    async def __await(self) -> dict[_K, _V]:
        result: dict[_K, _V] = {}
        exceptions: list[Exception] = []
        for key, coro in self.items():
            try:
                result[key] = await coro
            except Exception as exc:
                exceptions.append(exc)
        if exceptions:
            raise ExceptionGroup('Multiple exceptions occurred', exceptions)
        return result

    @classmethod
    def from_concurrent_futures(cls, futures: Mapping[_K, Future[_V]]) -> Self:
        return cls({key: wrap_concurrent_future(value) for key, value in futures.items()})


def to_list(obj: Iterable[_T]) -> list[_T]:
    if isinstance(obj, list):
        return obj
    return list(obj)


def to_dict(obj: Mapping[_K, _V]) -> dict[_K, _V]:
    if isinstance(obj, dict):
        return obj
    return dict(obj)


def to_set(obj: Iterable[_T]) -> set[_T]:
    if isinstance(obj, set):
        return obj
    return set(obj)


def make_kwargs(**kwargs: Any) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}


def make_sync_config(config: Mapping[str, Any]) -> dict[str, Any]:
    result = dict(config)
    for key in ['error_cb', 'throttle_cb', 'stats_cb', 'oauth_cb', 'on_delivery', 'on_commit']:
        if func := result.get(key):
            result[key] = async_to_sync(func)
    return result



import confluent_kafka
import confluent_kafka.admin

from kafka_async import Consumer, Producer


def test_assign(consumer: Consumer, topic: str) -> None:
    consumer.assign([confluent_kafka.TopicPartition(topic, 0)])
    assert consumer.assignment() == [confluent_kafka.TopicPartition(topic, 0)]


def test_commit__no_args(consumer: Consumer, topic: str) -> None:
    consumer.commit()


async def test_commit__sync(producer: Producer, consumer: Consumer, topic: str) -> None:
    producer.produce(topic=topic)
    await producer.flush(timeout=5)

    consumer.subscribe([topic])
    assert await consumer.consume(timeout=5)

    result = await consumer.commit(asynchronous=False)

    assert result == [confluent_kafka.TopicPartition(topic, 0, 1)]

    assert await consumer.committed([confluent_kafka.TopicPartition(topic, 0)]) == [
        confluent_kafka.TopicPartition(topic, 0, 1)
    ]


async def test_commit__by_message_sync(producer: Producer, consumer: Consumer, topic: str) -> None:
    producer.produce(topic=topic)
    await producer.flush(timeout=5)

    consumer.subscribe([topic])
    (msg,) = await consumer.consume(timeout=5)

    result = await consumer.commit(msg, asynchronous=False)

    assert result == [confluent_kafka.TopicPartition(topic, 0, 1)]


async def test_commit__by_offsets_sync(producer: Producer, consumer: Consumer, topic: str) -> None:
    producer.produce(topic=topic)
    await producer.flush(timeout=5)

    consumer.subscribe([topic])
    assert await consumer.consume(timeout=5)

    result = await consumer.commit(offsets=[confluent_kafka.TopicPartition(topic, 0, 1)], asynchronous=False)

    assert result == [confluent_kafka.TopicPartition(topic, 0, 1)]


async def test_get_watermark_offsets__cached(producer: Producer, consumer: Consumer, topic: str) -> None:
    producer.produce(topic=topic)
    await producer.flush(timeout=5)

    assert consumer.get_watermark_offsets(confluent_kafka.TopicPartition(topic, 0), cached=True) == (
        confluent_kafka.OFFSET_INVALID,
        confluent_kafka.OFFSET_INVALID,
    )


async def test_get_watermark_offsets__not_cached(producer: Producer, consumer: Consumer, topic: str) -> None:
    producer.produce(topic=topic)
    await producer.flush(timeout=5)

    assert await consumer.get_watermark_offsets(confluent_kafka.TopicPartition(topic, 0), cached=False) == (0, 1)


def test_incremental_assign(consumer: Consumer) -> None:
    consumer.incremental_assign([confluent_kafka.TopicPartition('topic', 0, 0)])
    consumer.incremental_unassign([confluent_kafka.TopicPartition('topic', 0, 0)])


async def test_list_topics(consumer: Consumer) -> None:
    metadata = await consumer.list_topics()
    assert isinstance(metadata, confluent_kafka.admin.ClusterMetadata)


async def test_memberid(consumer: Consumer, topic: str, producer: Producer) -> None:
    producer.produce(topic=topic)
    await producer.flush(timeout=5)

    consumer.subscribe([topic])
    await consumer.consume()
    memberid = consumer.memberid()
    assert memberid
    assert memberid.startswith('rdkafka')


def test_consumer_group_metadata(consumer: Consumer) -> None:
    assert consumer.consumer_group_metadata() is not None


def test_set_sasl_credentials(consumer: Consumer) -> None:
    consumer.set_sasl_credentials('user', 'password')
